from typing import Callable
from uuid import uuid4
from sqlmodel import Session
from kv.api import KV
from dslog import Logger
from moveread.core import CoreAPI, Game as CoreGame
from ...types import Game, GameId, Image, Pairing, Paired

def gameIdFn(game: CoreGame, tournId: str) -> GameId:
  if game.meta is None or game.meta.tournament is None:
    return GameId(tournId=tournId, group='a', round='1', board='1')
  t = game.meta.tournament
  return GameId(tournId=tournId, group=t.group or 'a', round=t.round or '1', board=t.board or '1')

async def input_core(
  core: CoreAPI, session: Session,
  *, images: KV[bytes],
  gameId_fn: Callable[[CoreGame], GameId] | None = None,
  tournId: str = 'llobregat23',
  num_games: int | None = None, shuffle: bool = True,
  logger = Logger.rich().prefix('[CORE INPUT]')
):
  """Input all images from `core` into `Qin` tasks
  - Actually, only images with `version == 0`
  - `model_fn`: determines the scoresheet model of each task
  - `state_fn`: determines an arbitrary tuple of JSON-serializable data to attach to each task
  """
  gameId_fn = gameId_fn or (lambda game: gameIdFn(game, tournId))
  games = list((await core.games.keys()).unsafe())
  if shuffle:
    import random
    random.shuffle(games)
  for gameId in games[:num_games]:
    game = (await core.games.read(gameId)).unsafe()
    game.id = f'{tournId}/{game.id}'
    imgs = []
    for imgId, image in game.images:
      if imgId.version == 0:
        id = str(imgId)
        url = f'{id}/original_{uuid4()}.jpg'
        img = (await core.blobs.read(image.url)).unsafe()
        (await images.insert(url, img)).unsafe()
        imgs.append(url)

    gid = gameId_fn(game)
    g = Game(
      index=int(gid['board']), pairing=Pairing(Paired(white='Magnus', black='Anish', tag='paired')),
      imgs=[Image(url=img, descaled_url=img) for img in imgs], status=Game.Status.uploaded, **gid)
    session.add(g)
    logger(f'Inputted game "{gid}"')
  session.commit()


from datetime import date
from sqlmodel import Session
from chess_pairings import PairingsSource, ChessResultsPairings
from moveread.dfy.types import Tournament
from ..pairings import create_tournament

ST_MARTI = Tournament(
  tournId='stmarti', name='XVIII Open Sant Martí', site='Barcelona',
  start_date=date(2023, 1, 1), end_date=date(2023, 1, 9), groups=['a', 'b']
)

ST_MARTI_PAIRINGS = dict(
  a=PairingsSource(ChessResultsPairings(db_key=741703)),
  b=PairingsSource(ChessResultsPairings(db_key=741704)),
)

def create_mock(session: Session):
  create_tournament(session, ST_MARTI, group_pairings=ST_MARTI_PAIRINGS, model='fcde')


from scoresheet_models import MODEL_IDS
from argparse import ArgumentParser

def main():
  parser = ArgumentParser()
  parser.add_argument('-i', '--input', required=True)
  parser.add_argument('-o', '--output', required=True)
  parser.add_argument('-b', '--base-path', required=True)
  parser.add_argument('-d', '--db', required=True, help='Database URL')
  parser.add_argument('-t', '--tournament', required=True, help='Tournament ID')
  parser.add_argument('-m', '--model', required=True, choices=MODEL_IDS, help='Model ID')
  parser.add_argument('--protocol', default='sqlite', required=False, choices=['sqlite', 'fs'], help='Protocol used in queues')
  parser.add_argument('--images', type=str, help='Local path to images', required=True)
  parser.add_argument('--blobs', type=str, help='Azure Blob connection string', required=True)
  parser.add_argument('--core', type=str, help='Path to output core', required=True)

  args = parser.parse_args()


  import os
  from dslog import Logger
  input_path = os.path.join(os.getcwd(), args.input)
  output_path = os.path.join(os.getcwd(), args.output)
  base_path = os.path.join(os.getcwd(), args.base_path)
  proto = args.protocol
  db_url = args.db
  tournId = args.tournament
  images_path = os.path.join(os.getcwd(), args.images)
  core_path = os.path.join(os.getcwd(), args.core)
  
  logger = Logger.click().prefix('[DFY]')
  logger(f'Running...')
  logger(f'- Tournament ID: "{tournId}"')
  logger(f'- Database URL: "{db_url}"')
  logger(f'- Images path: "{images_path}"')
  logger(f'- Output core path: "{core_path}"')
  logger(f'- Model: "{args.model}"')

  
  logger(f'- Queues protocol: "{proto}"')
  logger(f'- Input path: "{input_path}"')
  logger(f'- Internal path: "{base_path}"')
  logger(f'- Output path: "{output_path}"')
  

  if args.blobs.startswith('.'):
    from kv.fs import FilesystemKV
    blobs = os.path.join(os.getcwd(), args.blobs)
    online_images = FilesystemKV[bytes](blobs)
    logger(f'- Local blobs: {blobs}')
  
  else:
    from kv.azure.blob import BlobKV
    online_images = BlobKV[bytes].from_conn_str(args.blobs)
    logger(f'- Azure Blob account: {online_images.client.account_name}')
  
  from sqlmodel import create_engine
  from kv.fs import FilesystemKV
  from moveread.dfy import run_connect
  from moveread.pipelines.dfy import run_local, input_queue, output_queue
  from multiprocessing import Process
  from moveread.core import CoreAPI
  core = CoreAPI.at(core_path)
  local_images = FilesystemKV[bytes](images_path)

  Qin = input_queue(input_path, protocol=proto)
  Qout = output_queue(output_path, protocol=proto)
  engine = create_engine(db_url)
  ps = (
    Process(
      target=run_connect, args=(Qin, Qout), kwargs=dict(
        engine=engine, tournId=tournId, logger=logger.prefix('[I/O]'), model=args.model,
        local_images=local_images, online_images=online_images, output_core=core
      )
    ),
    Process(
      target=run_local, args=(Qin, Qout), kwargs=dict(
        base_path=base_path, images=local_images,
        images_path=images_path, logger=logger
      )
    ),
  )
  for p in ps:
    p.start()
  for p in ps:
    p.join()

if __name__ == '__main__':
  import sys
  import os
  from dotenv import load_dotenv
  path = '/home/m4rs/mr-github/modes/moveread-dfy/infra/doer/local'
  os.makedirs(path, exist_ok=True)
  os.chdir(path)
  load_dotenv()

  SQL_CONN_STR = os.environ['SQL_CONN_STR']
  # SQL_CONN_STR = 'sqlite:////home/m4rs/mr-github/modes/moveread-dfy/local-db/db.sqlite'
  BLOB_CONN_STR = os.environ['BLOB_CONN_STR']
  # BLOB_CONN_STR = '../../../local-db/images'
  
  args = f'-i queues/in -o queues/out -b queues/internal \
      -t llobregat23 -m llobregat23 \
      --db {SQL_CONN_STR} \
      --blobs {BLOB_CONN_STR} \
      --core core \
      --images images'.replace('\n', '').split(' ')
  args = [arg for arg in args if arg]
  sys.argv.extend(args)
  main()


from typing import Sequence, Literal, TypeAlias, Unpack, Any, Iterable
import asyncio
from sqlalchemy import Engine
from sqlmodel import Session, select
from uuid import uuid4
from haskellian import either as E, Left
import pure_cv as vc
from kv.api import KV, InexistentItem, ReadError
from chess_pairings import GameId, GroupId, RoundId
from .pgns import export_all
from ..types import Game, Token, Group, Image, Tournament, FrontendPGN
from ..lib import queries

ImgExtension: TypeAlias = Literal['.jpg', '.png', '.webp']
MimeType: TypeAlias = str

def stringify(tournId: str, group: str, round: str, board: str) -> str:
  return f'{tournId}/{group}/{round}/{board}'

class DFY:

  def __init__(self, images: KV[bytes], engine: Engine):
    self._images = images
    self._engine = engine

  def authorize(self, token: str, tournId: str) -> bool:
    with Session(self._engine) as ses:
      results = ses.exec(select(Token).where(Token.token == token, Token.tournId == tournId))
      return results.first() is not None

  def tournaments(self) -> Sequence[Tournament]:
    with Session(self._engine) as ses:
      return ses.exec(select(Tournament)).all()

  def tournament(self, tournId: str) -> Tournament | None:
    with Session(self._engine) as ses:
      return ses.exec(queries.select_tnmt(tournId)).first()
    
  def group(self, tournId: str, group: str) -> Group | None:
    with Session(self._engine) as ses:
      return ses.exec(select(Group).where(Group.tournId == tournId, Group.name == group)).first()
  
  def round(self, **roundId: Unpack[RoundId]) -> Sequence[Game]:
    with Session(self._engine) as ses:
      order: Any = Game.index # order_by's typing is messed up
      stmt = select(Game).where(*queries.round_games(**roundId)).order_by(order)
      return ses.exec(stmt).all()
    
  def round_pgn(self, **roundId: Unpack[RoundId]) -> Iterable[str]:
    with Session(self._engine) as ses:
      order: Any = Game.index
      round = ses.exec(queries.select_round(**roundId)).first()
      stmt = select(Game).where(*queries.round_games(**roundId)).order_by(order)
      games = ses.exec(stmt).all()
      tnmt = ses.exec(queries.select_tnmt(roundId['tournId'])).first()
      yield from export_all(games, tnmt, round)

  def group_pgn(self, **groupId: Unpack[GroupId]) -> Iterable[str]:
    with Session(self._engine) as ses:
      order: Any = Game.index
      tnmt = ses.exec(queries.select_tnmt(groupId['tournId'])).first()
      rounds = ses.exec(queries.select_rounds(**groupId)).all()
      for round in rounds:
        stmt = select(Game).where(*queries.round_games(round=round.name, **groupId)).order_by(order)
        games = ses.exec(stmt).all()
        yield from export_all(games, tnmt, round)

  def game_pgn(self, **gameId: Unpack[GameId]) -> FrontendPGN | None:
    with Session(self._engine) as ses:
      game = ses.exec(queries.select_game(**gameId)).first()
      if game and game.pgn:
        return game.pgn
      
  def images(self, **gameId: Unpack[GameId]) -> Sequence[str] | None:
    with Session(self._engine) as ses:
      game = ses.exec(queries.select_game(**gameId)).first()
      if game:
        return [img.descaled_url for img in game.imgs]
      
  @E.do[ReadError]()
  async def post_game(
    self, images: Sequence[bytes],
    descaled_height: int = 768,
    **gameId: Unpack[GameId]
  ):
    if images == []:
      return Left('No images provided')

    id = f'{stringify(**gameId)}_{uuid4()}'

    uploads = []
    imgs = []
    for i, img in enumerate(images):
      mat = vc.decode(img)
      url = f'{id}/{i}.jpg'
      descaled_url = f'{id}/{i}-{descaled_height}.jpg'
      descaled = vc.encode(vc.descale_h(mat, descaled_height), format='.jpg')
      encoded = vc.encode(mat, format='.jpg')
      uploads.extend([
        self._images.insert(url, encoded),
        self._images.insert(descaled_url, descaled)
      ])
      imgs.append(Image(url=url, descaled_url=descaled_url))
    
    E.sequence(await asyncio.gather(*uploads)).unsafe()

    deletions = []
    with Session(self._engine) as ses:
      game = ses.exec(queries.select_game(**gameId)).first()
      if game is None:
        return Left(InexistentItem(detail=f'Game {gameId} not found in DB'))
      for img in game.imgs:
        (await self._images.delete(img.url)).unsafe()
        ses.delete(img)
        deletions.extend([
          self._images.delete(img.url),
          self._images.delete(img.descaled_url)
        ])

      game.imgs = imgs
      game.status = Game.Status.uploaded
      ses.add(game)
      ses.commit()

    E.sequence(await asyncio.gather(*deletions)) # if they don't get deleted, who cares


from argparse import ArgumentParser
from scoresheet_models import MODEL_IDS

def main():
  parser = ArgumentParser()
  parser.add_argument('-i', '--input', required=True)
  parser.add_argument('-o', '--output', required=True)
  parser.add_argument('-d', '--database', required=True, help='Database URL')
  parser.add_argument('-t', '--tournament', required=True, help='Tournament ID')
  parser.add_argument('-m', '--model', required=True, choices=MODEL_IDS, help='Model ID')
  parser.add_argument('--protocol', default='sqlite', required=False, choices=['sqlite', 'fs'], help='Protocol used in queues')

  args = parser.parse_args()


  import os
  from dslog import Logger
  input_path = os.path.join(os.getcwd(), args.input)
  output_path = os.path.join(os.getcwd(), args.output)
  proto = args.protocol
  db_url = args.database
  tournId = args.tournament

  logger = Logger.rich().prefix('[DFY CONNECTOR]')
  
  logger(f'Running...')
  logger(f'- Tournament ID: "{tournId}"')
  logger(f'- Database URL: "{db_url}"')
  logger(f'- Queues protocol: "{proto}"')
  logger(f'- Input path: "{input_path}"')
  logger(f'- Output path: "{output_path}"')
  
  from sqlmodel import create_engine
  from moveread.dfy import run_connect
  from moveread.pipelines.dfy import input_queue, output_queue
  
  Qin = input_queue(input_path, protocol=proto)
  Qout = output_queue(output_path, protocol=proto)
  engine = create_engine(db_url)
  run_connect(Qin, Qout, engine=engine, tournId=tournId, logger=logger, model=args.model)

if __name__ == '__main__':
  import sys
  import os
  os.chdir('/home/m4rs/mr-github/modes/dfy/moveread-dfy/')
  sys.argv.extend('-i queues/in -o queues/out -d sqlite:///db.sqlite -t llobregat -m llobregat23'.split(' '))
  main()


import os
from argparse import ArgumentParser
from openapi_ts import generate_client
from moveread.dfy import fastapi

def main():

  parser = ArgumentParser()
  parser.add_argument('-p', '--package-path', help="Path to the typescript package's base folder", required=True)
  args = parser.parse_args()

  app = fastapi({}, blobs={}) # type: ignore
  spec = app.openapi()
  path = os.path.abspath(args.package_path)
  generate_client(spec, path)

if __name__ == '__main__':
  main()


import asyncio
from sqlalchemy import Engine
from sqlmodel import select, Session
from dslog import Logger
from haskellian import either as E, Left
from kv.api import KV
from q.api import ReadQueue
from moveread.core import CoreAPI
from moveread.pipelines.dfy import Result, output_one
from ..types import Game, PGN

def exact_game(tournId: str, group: str, round: str, board: str):
  return Game.tournId == tournId, Game.group == group, Game.round == round, Game.board == board

async def pusher(
  Qout: ReadQueue[Result], engine: Engine, *,
  output_core: CoreAPI, images: KV[bytes],
  logger = Logger.rich().prefix('[DFY PUSHER]')
):
  
  @E.do()
  async def push_one():
    id, result = (await Qout.read()).unsafe()
    (await output_one(output_core, id, result, images=images)).unsafe()
    gid = result.gameId
    logger(f'Pushing result for {gid}')
    try:
      with Session(engine) as ses:
        stmt = select(Game).where(*exact_game(**gid))
        game = ses.exec(stmt).first()
        if game is None:
          logger(f'Game not found: {gid}', level='ERROR')
          return

        game.pgn = PGN(moves=result.pgn, early=result.early)
        game.status = Game.Status.done
        ses.add(game)
        ses.commit()
    except Exception as e:
      Left(e).unsafe()

    (await Qout.pop(id)).unsafe()

  while True:
    r = await push_one()
    if r.tag == 'left':
      logger('Error whilst pushing', r.value, level='ERROR')
      await asyncio.sleep(5)


from .queries import exact_game, exact_round, group_games, round_games, \
  select_game, select_round, select_rounds, select_tnmt, select_group_games

__all__ = [
  'exact_game', 'exact_round', 'group_games', 'round_games',
  'select_game', 'select_round', 'select_rounds', 'select_tnmt',
  'select_group_games',
]


from .types import Game, GameId, Image, Tournament, PGN, Pairing
from .server import DFY, fastapi, run_api
from .doer import run_connect
from .integrations import input_core

__all__ = [
  'Game', 'GameId', 'Image', 'Tournament', 'PGN', 'Pairing',
  'DFY', 'fastapi', 'run_api',
  'run_connect', 'input_core'
]


from argparse import ArgumentParser
import os

def env(variable: str, *, default = None, required: bool = True) -> dict:
  if (value := os.getenv(variable, default)) is not None:
    return dict(default=value)
  return dict(required=required)

def main():
  parser = ArgumentParser()
  parser.add_argument('--dfy-db', **env('DFY_DB'), help='Database URL')
  parser.add_argument('--dfy-images', **env('DFY_IMAGES'), type=str, help='KV conn str to DFY images')
  parser.add_argument('--doer-db', **env('DOER_DB'), help='Internal database URL, for queues and stuff')
  parser.add_argument('--doer-images', **env('DOER_IMAGES'), type=str, help='KV conn str to doer images')
  parser.add_argument('--core-games', **env('CORE_GAMES'), type=str, help='KV conn str to output Core games')
  parser.add_argument('--core-blobs', **env('CORE_BLOBS'), type=str, help='KV conn str to output Core blobs')
  parser.add_argument('--tf-host', **env('TF_HOST', required=False), type=str)
  parser.add_argument('--tf-port', **env('TF_PORT', required=False), type=str)

  parser.add_argument('-p', '--port', default=8000, type=int)
  parser.add_argument('--host', default='0.0.0.0', type=str)

  args = parser.parse_args()

  from dslog import Logger
  dfy_db = args.dfy_db
  doer_db = args.doer_db
  dfy_images = args.dfy_images
  doer_images = args.doer_images
  core_games = args.core_games
  core_blobs = args.core_blobs
  
  logger = Logger.click().prefix('[DFY]')
  logger(f'Running...')
  logger(f'- Database URL: "{dfy_db}"')
  logger(f'- Internal DB URL: "{doer_db}"')
  logger(f'- Doer images URL: "{doer_images}"')
  logger(f'- DFY images URL: "{dfy_images}"')
  logger(f'- Core games URL: "{core_games}"')
  logger(f'- Core blobs URL: "{core_blobs}"')

  from typing import Sequence
  from functools import cache
  import asyncio
  from multiprocessing import Process
  import uvicorn
  from sqlmodel import create_engine
  from kv.api import KV
  from kv.fs import FilesystemKV
  from kv.sql import SQLKV
  from q.kv import QueueKV
  import tf.serving as tfs
  from moveread.core import CoreAPI
  import moveread.pipelines.preprocess as pre
  from moveread.pipelines.game_preprocess import Game
  from moveread.pipelines.dfy import Workflow, Result
  from moveread.dfy.doer.main import artifacts
  
  online_engine = create_engine(dfy_db)
  engine = create_engine(doer_db)

  @cache
  def make_queue(path: Sequence[str], type: type):
    kv = SQLKV(type, engine, table='-'.join(['queues', *path]))
    return QueueKV(kv)
  
  Qout = make_queue(('output',), Result)
  queues = Workflow.make_queues(make_queue, Qout)

  online_images = KV.of(dfy_images)
  local_images = KV.of(doer_images)
  if isinstance(local_images, FilesystemKV):
    images_path = local_images.base_path
  else:
    images_path = None

  core = CoreAPI.of(core_games, core_blobs)

  tfserving = tfs.Params()
  if args.tf_host: tfserving['host'] = args.tf_host
  if args.tf_port: tfserving['port'] = args.tf_port
  
  artifs = artifacts(Qout=Qout, **queues)(
    online_images=online_images,
    images=local_images, images_path=images_path,
    games=SQLKV(Game, engine, 'games'),
    imgGameIds=SQLKV(str, engine, 'game-ids'),
    received_imgs=SQLKV(dict[str, pre.Result], engine, 'received-imgs'),
    engine=online_engine, output_core=core,
    tfserving=tfserving
  )

  ps = {
    id: Process(target=asyncio.run, args=(f,)) for id, f in artifs.processes.items()
  } | {
    'api': Process(target=uvicorn.run, args=(artifs.api,), kwargs={'host': args.host, 'port': args.port})
  }
  for id, p in ps.items():
    p.start()
    logger(f'Process "{id}" started at PID {p.pid}')
  for p in ps.values():
    p.join()

if __name__ == '__main__':
  import sys
  import os
  from dotenv import load_dotenv
  load_dotenv()

  def params(mock: bool = True):
    if mock:
      path = '/home/m4rs/mr-github/modes/moveread-dfy/infra/doer/local'
      os.makedirs(path, exist_ok=True)
      os.chdir(path)
      os.makedirs('core', exist_ok=True)
      return {
        'dfy-db': 'sqlite:///../../local-db/db.sqlite',
        'dfy-images': 'file://../../local-db/images',
        'doer-images': 'file://images',
        'doer-db': 'sqlite:///db.sqlite',
        'core-games': 'sql+sqlite:///core/games.sqlite;table=games',
        'core-blobs': 'file://core/images',
      }
    else:
      return {
        'dfy-db': os.getenv('DFY_DB'),
        'dfy-images': os.getenv('DFY_IMAGES'),
        'doer-images': os.getenv('DOER_IMAGES'),
        'doer-db': os.getenv('DOER_DB'),
        'core-games': os.getenv('CORE_GAMES'),
        'core-blobs': os.getenv('CORE_BLOBS'),
      }

  args = []
  for k, v in params().items():
    args.append(f'--{k}')
    args.append(v)
  print(args)
  sys.argv.extend(args)
  main()


from typing import Unpack
from sqlmodel import select
from chess_pairings import GameId, RoundId
from moveread.dfy.types import Game, Tournament, Round

def group_games(tournId: str, group: str):
  return Game.tournId == tournId, Game.group == group

def round_games(tournId: str, group: str, round: str):
  return Game.tournId == tournId, Game.group == group, Game.round == round

def exact_game(tournId: str, group: str, round: str, board: str):
  return Game.tournId == tournId, Game.group == group, Game.round == round, Game.board == board

def exact_round(tournId: str, group: str, round: str):
  return Round.tournId == tournId, Round.group == group, Round.name == round

def select_game(**gameId: Unpack[GameId]):
  return select(Game).where(*exact_game(**gameId))

def select_group_games(tournId: str, group: str):
  return select(Game).where(*group_games(tournId, group))

def select_tnmt(tournId: str):
  return select(Tournament).where(Tournament.tournId == tournId)

def select_round(**roundId: Unpack[RoundId]):
  return select(Round).where(*exact_round(**roundId))

def select_rounds(tournId: str, group: str):
  return select(Round).where(Round.tournId == tournId, Round.group == group)



from typing import Sequence
from sqlmodel import Session, select
from moveread.dfy.types import Token

def list(session: Session, tournId: str) -> Sequence[str]:
  return session.exec(select(Token.token).where(Token.tournId == tournId)).all()

def delete(session: Session, tournId: str, token: str):
  stmt = select(Token).where(Token.tournId == tournId, Token.token == token)
  obj = session.exec(stmt).first()
  if obj:
    session.delete(obj)
    session.commit()

def create(session: Session, tournId: str, token: str):
  obj = Token(tournId=tournId, token=token)
  session.add(obj)
  session.commit()



from typing import Sequence, Mapping, Literal, Any
from haskellian import Either, Left, Right
from sqlmodel import Session, select
import chess_pairings as cp
from scoresheet_models import ModelID
from moveread.dfy.types import Tournament, Group, Pairings, SheetModel, Game, Pairing
from . import queries

def create_tournament(
  session: Session, tnmt: Tournament, *,
  group_pairings: Mapping[str, cp.PairingsSource] = {},
  model: ModelID | None = None,
):
  for name in tnmt.groups:
    group = Group(tournId=tnmt.tournId, name=name, rounds=[])
    session.add(group)

  for group, pairings in group_pairings.items():
    session.add(Pairings(tournId=tnmt.tournId, group=group, pairings=pairings))

  if model:
    session.add(SheetModel(tournId=tnmt.tournId, model=model))

  session.add(tnmt)
  session.commit()
  


def insert_pairings(session: Session, *, tournId: str, group: str, pairings: cp.GroupPairings) -> Sequence[cp.GameId]:
    """Inserts pairings to games that didn't already exist"""
    group_games = session.exec(queries.select_group_games(tournId, group)).all()
    games_idx = cp.GamesMapping.from_pairs([(g.gameId(), g) for g in group_games])
    added: list[cp.GameId] = []

    for round, rnd_pairings in pairings.items():
      for board, pair in rnd_pairings.items():
        gid = cp.gameId(tournId, group, round, board)
        if not gid in games_idx:
          game = Game(tournId=tournId, group=group, round=round, board=board, index=int(board)-1, pairing=Pairing(pair))
          session.add(game)
          added.append(gid)
    
    session.commit()
    return added



async def update_pairings(session: Session, tournId: str) -> Either[Literal['no-pairings'], Mapping[str, Either[Any, Sequence[cp.GameId]]]]:
  """Inserts pairings if inexistent. If successful, returns a sequence added games"""
  pairings = session.exec(select(Pairings).where(Pairings.tournId == tournId)).all()
  
  if pairings == []:
    return Left('no-pairings')
  
  results = {}
  for p in pairings:
    pairs = await cp.scrape_pairings(p.pairings)
    if pairs.tag == 'left':
      results[p.group] = pairs
    else:
      added = insert_pairings(session, tournId=tournId, group=p.group, pairings=pairs.value)
      results[p.group] = Right(added)
    
  return Right(results)



from .mock import tournament
from .pairings import create_tournament, update_pairings, insert_pairings
from .misc import update_pgns
from . import queries, tokens, mock

__all__ = [
  'create_tournament', 'update_pairings', 'insert_pairings', 'update_pgns',
  'queries', 'tournament', 'tokens', 'mock',
]


import os
import uvicorn
from dslog import Logger
from kv.api import LocatableKV
from moveread.dfy import fastapi, DFY
from sqlalchemy import Engine

def run_api(
  images: LocatableKV[bytes],
  engine: Engine, *,
  images_path: str | None = None,
  port: int = 8000,
  host: str = '0.0.0.0',
  logger = Logger.click().prefix('[DFY API]')
):
  if images_path is not None:
    os.makedirs(images_path, exist_ok=True)
  sdk = DFY(images=images, engine=engine)
  app = fastapi(sdk, blobs=images, images_path=images_path, logger=logger)
  uvicorn.run(app, port=port, host=host)


from typing import Unpack
import asyncio
from datetime import timedelta
from sqlalchemy import Engine
from dslog import Logger
from kv.api import KV
from q.api import ReadQueue
from scoresheet_models import ModelID
from moveread.core import CoreAPI
import moveread.pipelines.dfy as dfy
from ._puller import puller
from ._pusher import pusher

def artifacts(*, Qout: ReadQueue[dfy.Result], **queues: Unpack[dfy.Workflow.Queues]):
  
  def _bound(
      logger = Logger.click().prefix('[DFY]'), *,
      engine: Engine, output_core: CoreAPI,
      online_images: KV[bytes], polling_interval = timedelta(seconds=30),
      **params: Unpack[dfy.Params]
    ) -> dfy.Artifacts:
    
    async def coro():
      await asyncio.gather(
        puller(
          queues['Qin'], engine, polling_interval=polling_interval,
          logger=logger.prefix('[PULLER]'),
          online_images=online_images, local_images=params['images']
        ),
        pusher(Qout, engine, output_core=output_core, images=params['images'], logger=logger.prefix('[PUSHER]'))
      )

    artifs = dfy.Workflow.artifacts(**queues['internal'])(logger=logger, **params)
    artifs.processes = { 'connect': coro() } | {
      f'doer-{id}': proc for id, proc in artifs.processes.items()
    }
    
    return artifs
  
  return _bound


import os
from argparse import ArgumentParser

def main(): 
  parser = ArgumentParser()
  parser.add_argument('--images', type=str, help='KV connection string to images')
  parser.add_argument('--db', required=True, type=str, help='DB URL')

  parser.add_argument('-p', '--port', default=8000, type=int)
  parser.add_argument('--host', default='0.0.0.0', type=str)


  args = parser.parse_args()
  db = args.db

  from dslog import Logger, formatters
  logger = Logger.stderr().format(formatters.click)

  logger(f'Running API...')
  logger(f'- DB URL: "{db.split(":")[0]}://[...]{db.split(":")[1][-16:]}"')
  logger(f'- Images URL: {args.images.split(":")[0]}://[...]{args.images.split(":")[1][-16:]}')

  from kv.api import KV, LocatableKV
  from sqlalchemy import create_engine
  from moveread.dfy import run_api

  images = KV[bytes].of(args.images)
  images_path = getattr(images, 'base_path', None)
  if not isinstance(images, LocatableKV):
    raise ValueError('Images KV must be locatable')

  engine = create_engine(db)
  run_api(images, engine, images_path=images_path, port=args.port, host=args.host, logger=logger)

if __name__ == '__main__':
  import os
  import sys
  from dotenv import load_dotenv
  load_dotenv()

  MOCK = True
  if MOCK:
    os.chdir('/home/m4rs/mr-github/modes/moveread-dfy/infra/local-db')
    SQL_CONN_STR = 'sqlite:///db.sqlite'
    IMAGES = 'file:///home/m4rs/mr-github/modes/moveread-dfy/local-db/images'
    sys.argv.extend(f'--images {IMAGES} --db {SQL_CONN_STR}'.split(' '))
  else:
    SQL_CONN_STR = os.environ['SQL_CONN_STR']
    BLOB_CONN_STR = os.environ['BLOB_CONN_STR']
    sys.argv.extend(f'--images {BLOB_CONN_STR} --db {SQL_CONN_STR}'.split(' '))
  
  main()


from typing import Sequence, Literal, Annotated
import os
from datetime import datetime, timedelta
from fastapi import FastAPI, Response, File, Request, status
from fastapi.responses import StreamingResponse
from fastapi.staticfiles import StaticFiles
from fastapi.middleware.cors import CORSMiddleware
from dslog import Logger
from dslog.uvicorn import setup_loggers_lifespan, DEFAULT_FORMATTER, ACCESS_FORMATTER
from kv.api import LocatableKV
from chess_pairings import roundId, gameId, groupId
from ..types import FrontendGame, Tournament, Group, FrontendPGN
from .sdk import DFY

def fastapi(
  sdk: DFY, *, images_path: str | None = None,
  blobs: LocatableKV[bytes],
  logger = Logger.click().prefix('[DFY API]')
):
  
  app = FastAPI(
  generate_unique_id_function=lambda route: route.name,
    lifespan=setup_loggers_lifespan(
      access=logger.format(ACCESS_FORMATTER),
      uvicorn=logger.format(DEFAULT_FORMATTER),
    )
  )

  origins = [
    'http://localhost:5173',
    'http://localhost:4713',
    'https://moveread.com',
    'https://dfy.moveread.com',
  ]
  
  app.add_middleware(
    CORSMiddleware,
    allow_origins=origins,
    allow_credentials=True,
    allow_methods=["*"],
    allow_headers=["*"],
  )

  if images_path is not None:
    app.mount('/images', StaticFiles(directory=images_path))

  @app.get('/authorize/{tournId}', responses={ 401: {}, 200: {} })
  def authorize(token: str, tournId: str):
    authed = sdk.authorize(token, tournId)
    return Response(status_code=200 if authed else status.HTTP_401_UNAUTHORIZED)

  @app.get('/')
  def tournaments() -> Sequence[Tournament]:
    return sdk.tournaments()

  @app.get('/{tournId}', responses={ 404: {}, 200: { 'model': Tournament }})
  def tournament(tournId: str, r: Response) -> Tournament | None:
    tnmt = sdk.tournament(tournId)
    if tnmt is None:
      r.status_code = status.HTTP_404_NOT_FOUND
    return tnmt
  
  @app.get('/{tournId}/{group}.pgn', response_model=str)
  def group_pgn(tournId: str, group: str):
    pgn = sdk.group_pgn(**groupId(tournId, group))
    return StreamingResponse(content=pgn, media_type='application/x-chess-pgn', headers={
      'Content-Disposition': f'attachment; filename={tournId}_{group}.pgn'
    })
  
  @app.get('/{tournId}/{group}', responses={ 404: {}, 200: { 'model': Group }})
  def group(tournId: str, group: str, r: Response) -> Group | None:
    grp = sdk.group(tournId, group)
    if grp is None:
      r.status_code = status.HTTP_404_NOT_FOUND
    return grp

  @app.get('/{tournId}/{group}/{round}.pgn')
  def round_pgn(tournId: str, group: str, round: str):
    pgn = sdk.round_pgn(**roundId(tournId, group, round))
    return StreamingResponse(content=pgn, media_type='application/x-chess-pgn', headers={
      'Content-Disposition': f'attachment; filename={tournId}_{group}_{round}.pgn'
    })
  
  @app.get('/{tournId}/{group}/{round}')
  def round(tournId: str, group: str, round: str) -> Sequence[FrontendGame]:
    return sdk.round(**roundId(tournId, group, round))
  

  @app.get('/{tournId}/{group}/{round}/{board}/pgn', responses={ 404: {}, 200: { 'model': FrontendPGN }})
  def game_pgn(tournId: str, group: str, round: str, board: str, r: Response) -> FrontendPGN | None:
    pgn = sdk.game_pgn(**gameId(tournId, group, round, board))
    if pgn is None:
      r.status_code = status.HTTP_404_NOT_FOUND
    return pgn
  
  @app.get('/{tournId}/{group}/{round}/{board}/images', responses={ 404: {}, 200: { 'model': Sequence[str] }})
  def images(tournId: str, group: str, round: str, board: str, req: Request, res: Response) -> Sequence[str] | None:
    urls = sdk.images(**gameId(tournId, group, round, board))
    if urls is None or len(urls) == 0:
      res.status_code = status.HTTP_404_NOT_FOUND
      return None
    
    if images_path is not None:
      path = os.path.join(str(req.base_url), 'images')
      return [os.path.join(path, url) for url in urls]
    else:
      expiry = datetime.now() + timedelta(hours=1)
      return [blobs.url(url, expiry=expiry) for url in urls]
  
  @app.post('/{tournId}/{group}/{round}/{board}', responses={
    401: dict(model=Literal['UNAUTHORIZED']),
    200: dict(model=Literal['OK']),
    500: dict(model=Literal['ERROR'])
  })
  async def post_game(
    tournId: str, group: str, round: str, board: str, token: str,
    images: Annotated[list[bytes], File()], r: Response, 
  ) -> Literal['OK', 'UNAUTHORIZED', 'ERROR']:
  
    if not sdk.authorize(token, tournId):
      r.status_code = status.HTTP_401_UNAUTHORIZED
      return 'UNAUTHORIZED'
    
    gid = gameId(tournId, group, round, board)
    res = await sdk.post_game(images, **gid)
    if res.tag == 'left':
      logger(f'Error posting game {gid}:', res.value, level='ERROR')
      r.status_code = status.HTTP_500_INTERNAL_SERVER_ERROR
      return 'ERROR'
    
    return 'OK'  
  
  return app


import os
from argparse import ArgumentParser

def main(): 
  parser = ArgumentParser()
  group = parser.add_mutually_exclusive_group()
  group.add_argument('--images', type=str, help='Local path to images')
  group.add_argument('--blobs', type=str, help='Azure Blob connection string')

  parser.add_argument('--db', required=True, type=str, help='DB URL')

  parser.add_argument('-p', '--port', default=8000, type=int)
  parser.add_argument('--host', default='0.0.0.0', type=str)


  args = parser.parse_args()
  db = args.db

  from dslog import Logger, formatters
  logger = Logger.stderr().format(formatters.click)

  logger(f'Running API...')
  logger(f'- DB URL: "{db}"')

  if args.images:
    images_path = os.path.join(os.getcwd(), args.images)
    logger(f'- Images path: "{images_path}"')
    from kv.fs import FilesystemKV
    images = FilesystemKV[bytes](images_path)
  else:
    images_path = None
    from kv.azure.blob import BlobKV
    images = BlobKV[bytes].from_conn_str(args.blobs)
    logger(f'- Azure Blob account: {images.client.account_name}')

  from sqlalchemy import create_engine
  from moveread.dfy import run_api
  engine = create_engine(db)
  run_api(images, engine, images_path=images_path, port=args.port, host=args.host, logger=logger)

if __name__ == '__main__':
  import sys
  from dotenv import load_dotenv
  load_dotenv()

  MOCK = False
  if MOCK:
    SQL_CONN_STR = 'sqlite:////home/m4rs/mr-github/modes/moveread-dfy/local-db/db.sqlite'
    IMAGES = '/home/m4rs/mr-github/modes/moveread-dfy/local-db/images'
    sys.argv.extend(f'--images {IMAGES} --db {SQL_CONN_STR}'.split(' '))
  else:
    SQL_CONN_STR = os.environ['SQL_CONN_STR']
    BLOB_CONN_STR = os.environ['BLOB_CONN_STR']
    sys.argv.extend(f'--blobs {BLOB_CONN_STR} --db {SQL_CONN_STR}'.split(' '))
  
  main()


from typing import Sequence
from enum import StrEnum
from datetime import date, datetime
from pydantic import RootModel, Field as PydanticField
from sqlmodel import Field, SQLModel, Relationship
from sqltypes import SpaceDelimitedList, PydanticModel, ValidatedLiteral
from chess_pairings import Paired, Unpaired, GameId, gameId, PairingsSource
from scoresheet_models import ModelID

class Pairing(RootModel):
  root: Paired | Unpaired = PydanticField(discriminator='tag')

class Tournament(SQLModel, table=True):
  tournId: str = Field(primary_key=True)
  name: str
  site: str | None = None
  start_date: date | None = None
  end_date: date | None = None
  groups: Sequence[str] = Field(sa_type=SpaceDelimitedList)

class SheetModel(SQLModel, table=True):
  tournId: str = Field(primary_key=True, foreign_key='tournament.tournId')
  model: ModelID = Field(sa_type=ValidatedLiteral(ModelID))

class Pairings(SQLModel, table=True):
  tournId: str = Field(primary_key=True, foreign_key='tournament.tournId')
  group: str = Field(primary_key=True, foreign_key='group.name')
  pairings: PairingsSource = Field(sa_type=PydanticModel(PairingsSource))

class Group(SQLModel, table=True):
  tournId: str = Field(primary_key=True, foreign_key='tournament.tournId')
  name: str = Field(primary_key=True)
  rounds: Sequence[str] = Field(sa_type=SpaceDelimitedList)

class Round(SQLModel, table=True):
  tournId: str = Field(primary_key=True, foreign_key='tournament.tournId')
  group: str = Field(primary_key=True)
  name: str = Field(primary_key=True)
  start_dtime: datetime | None = None
  
class Image(SQLModel, table=True):
  id: int | None = Field(default=None, primary_key=True)
  url: str
  descaled_url: str
  gameId: int = Field(default=None, foreign_key='game.id')

class FrontendPGN(SQLModel):
  moves: Sequence[str] = Field(sa_type=SpaceDelimitedList)
  early: bool | None = None

class PGN(FrontendPGN, table=True):
  gameId: int = Field(default=None, primary_key=True, foreign_key='game.id')

class FrontendGame(SQLModel):
  board: str
  pairing: Pairing = Field(sa_type=PydanticModel(Pairing))
  status: 'Game.Status | None' = None

class Game(FrontendGame, table=True):
  class Status(StrEnum):
    uploaded = 'uploaded'
    doing = 'doing'
    done = 'done'

  id: int | None = Field(default=None, primary_key=True)
  tournId: str = Field(foreign_key='tournament.tournId')
  group: str
  round: str
  index: int
  """Boards may have out of whack names (e.g. in team tournaments "1.3"). This is the order you'd see in chess-results"""
  imgs: list[Image] = Relationship()
  pgn: PGN | None = Relationship()

  def gameId(self) -> GameId:
    return gameId(self.tournId, self.group, self.round, self.board)
  
class Token(SQLModel, table=True):
  id: int | None = Field(default=None, primary_key=True)
  token: str
  tournId: str
  
__all__ = [
  'Game', 'GameId', 'Image', 'Tournament', 'PGN', 'Pairing', 'Token'
]


from datetime import timedelta
import asyncio
from uuid import uuid4
from sqlalchemy import Engine
from sqlalchemy.exc import DatabaseError
from sqlmodel import select, Session
from haskellian import Left, either as E
from dslog import Logger
from kv.api import KV
from q.api import WriteQueue
from chess_pairings import gameId
from moveread.pipelines.dfy import Input
from ..types import Game, Pairing, Tournament, SheetModel

def randomId(tournId: str, group: str, round: str, board: str) -> str:
  return f'{tournId}/{group}/{round}/{board}_{uuid4()}'

def pairing_display(pairing: Pairing):
  pair = pairing.root
  if pair.tag == 'unpaired':
    return 'Unpaired!?'
  
  s = f'{pair.white} - {pair.black}'
  if pair.result is not None:
    s += f' {pair.result}'
  return s


def title(pairing: Pairing, tournId: str, group: str, round: str, board: str) -> str:
  return f'{tournId} {group}/{round}/{board} {pairing_display(pairing)}'

async def puller(
  Qin: WriteQueue[Input], engine: Engine, *,
  online_images: KV[bytes], local_images: KV[bytes],
  polling_interval: timedelta = timedelta(seconds=30),
  logger = Logger.rich().prefix('[PULLER]')
):
  @E.do()
  async def pull_once():
    try:
      with Session(engine) as ses:
        on = Game.tournId == SheetModel.tournId
        stmt = select(Game, SheetModel).join(SheetModel, on).where(Game.status == Game.Status.uploaded)
        for game, model in ses.exec(stmt):
          gid = gameId(game.tournId, game.group, game.round, game.board)
          id = randomId(**gid)
          urls = [img.url for img in game.imgs]
          task = Input(gameId=gid, model=model.model, imgs=urls, title=title(game.pairing, **gid))
          tasks = [online_images.copy(url, local_images, url).run() for url in urls]
          results = await asyncio.gather(*tasks)
          E.sequence(results).unsafe()

          logger(f'Inputting new task for "{id}":', task)
          (await Qin.push(id, task)).unsafe()
          game.status = Game.Status.doing
          ses.add(game)
        ses.commit()


    except DatabaseError as e:
      Left(e).unsafe()


  while True:
    logger('Polling', level='DEBUG')
    r = await pull_once()
    logger('Polled', level='DEBUG')
    if r.tag == 'left':
      logger('Error while pulling', r.value, level='ERROR')
    
    await asyncio.sleep(polling_interval.total_seconds())
